/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.NoDatabaseAssociativityException;
import org.ucam.srcf.assassins.exceptions.ObjectModifiedException;

/**
 * Class to represent canonical data attached to an assassin. Currently this
 * consists solely of the player by whom that assassin is played.
 * @see AssassinState
 *
 * @author Matthew Johnson <autoumpire@matthew.ath.cx>
 */
public class Assassin extends CanonicalEntity
{
   /** A Map of all currently instantiated Assassin objects */
   private static Map assassinCache = new HashMap();
   /** Objects which have been created and not committed */
   private static List newAssassinCache = new Vector();


   /** 
    * Gets a list of all the assassins from the database 
    *
    * @return A List of Assassin objects
    * 
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested aid wasn't found. 
    */
   static List getAllAssassins() throws AssassinSQLException
   {
      // get all the aids
      String query = "SELECT aid FROM " + DatabaseBackend.getPrefix() + "_assassins";
      AResultSet r = DatabaseBackend.query(query);
      
      List assassins = new Vector();
      try {
         while (r.encapsulated.next()) {
            int aid = r.encapsulated.getInt("aid");
            Integer AID = new Integer(aid);

            // check the cache, and add it if necessary
            if (assassinCache.containsKey(AID))
               assassins.add(assassinCache.get(AID));
            else {
               Assassin a = new Assassin(aid);
               assassins.add(a);
               assassinCache.put(AID, a);            
            }

         }
         r.close();
      } catch(SQLException SQLe) { throw new AssassinSQLException(SQLe); }      
      return assassins;
   }


   /** 
    * Gets a list of all the assassin viewers associated with a player 
    *
    * @return A List of AssassinViewer objects
    * 
    * @param player The player to search for
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested aid wasn't found. 
    */
   public static List getAssassinsOfPlayer(PlayerViewer player) throws AssassinSQLException
   {
      return getAssassinsOfPlayer((Player) player.getEntity());
   }
   /** 
    * Gets a list of all the assassin associated with a player 
    *
    * @return A List of Assassin objects
    * 
    * @param player The player to search for
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested aid wasn't found. 
    */
   static List getAssassinsOfPlayer(Player player) throws AssassinSQLException
   {
      // get all the aids
      String query = "SELECT a.aid, a.pid FROM " + DatabaseBackend.getPrefix() + 
           "_assassins a WHERE a.pid='"+player.getPid()+"'";
      AResultSet r = DatabaseBackend.query(query);
      
      List assassins = new Vector();
      try {
         while (r.encapsulated.next()) {
            int aid = r.encapsulated.getInt("aid");
            Integer AID = new Integer(aid);

            // check the cache, and add it if necessary
            if (assassinCache.containsKey(AID))
               assassins.add(assassinCache.get(AID));
            else {
               Assassin a = new Assassin(aid);
               assassins.add(a);
               assassinCache.put(AID, a);            
            }
         }
         r.close();
      } catch(SQLException SQLe) { throw new AssassinSQLException(SQLe); }      
      return assassins;
   }


   /** 
    * Gets a list of all the assassin viewers associated with a player with a given email
    *
    * @return A List of AssassinViewer objects
    * 
    * @param email The email to search for
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested aid wasn't found. 
    */
   public static List getAllAssassinViewersWithEmail(String email) throws AssassinSQLException
   {
      // get all the aids
      // select aid from m2004_copy_test_players natural join m2004_copy_test_assassins where email='mjj29@cam.ac.uk'
      String query = "SELECT a.aid, a.pid FROM " + DatabaseBackend.getPrefix() + 
           "_assassins a NATURAL JOIN " + DatabaseBackend.getPrefix() + "_players p WHERE p.email='"+email+"'";
      AResultSet r = DatabaseBackend.query(query);
      
      List assassins = new Vector();
      try {
         while (r.encapsulated.next()) {
            int aid = r.encapsulated.getInt("aid");
            Integer AID = new Integer(aid);

            // check the cache, and add it if necessary
            if (assassinCache.containsKey(AID))
               assassins.add(new AssassinViewer((Assassin) assassinCache.get(AID)));
            else {
               Assassin a = new Assassin(aid);
               assassins.add(new AssassinViewer(a));
               assassinCache.put(AID, a);            
            }
         }
         r.close();
      } catch(SQLException SQLe) { throw new AssassinSQLException(SQLe); }      
      return assassins;
   }


   /** 
    * Gets a list of all the assassins from the database sorted by Real Name
    *
    * @return A List of Assassin objects
    * 
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested aid wasn't found. 
    */
   static List getAllAssassinsOrderByRealName() throws AssassinSQLException
   {
      // get all the aids
      String query = "SELECT a.aid, a.pid FROM " + DatabaseBackend.getPrefix() + 
           "_assassins a NATURAL JOIN " + DatabaseBackend.getPrefix() + "_players p ORDER BY p.name";
      AResultSet r = DatabaseBackend.query(query);
      
      List assassins = new Vector();
      try {
         while (r.encapsulated.next()) {
            int aid = r.encapsulated.getInt("aid");
            Integer AID = new Integer(aid);

            // check the cache, and add it if necessary
            if (assassinCache.containsKey(AID))
               assassins.add(assassinCache.get(AID));
            else {
               Assassin a = new Assassin(aid);
               assassins.add(a);
               assassinCache.put(AID, a);            
            }

         }
         r.close();
      } catch(SQLException SQLe) { throw new AssassinSQLException(SQLe); }      
      return assassins;
   }

   /** 
    * Gets a list of all the assassins as Viewers
    *
    * @return A list of AssassinViewer objects
    *
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested aid wasn't found. 
    */
   public static List getAllAssassinViewers() throws AssassinSQLException
   {
      return Tools.viewerifyList(getAllAssassins());
   }
      
   /**
    * Creates a assassin by aid, checking existance in the database
    *
    * @param aid The database key aid to create from
    * @return The Assassin object corresponding to the given aid
    *
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested aid wasn't found. 
    */
   static Assassin getByAid(int aid) throws AssassinSQLException
   {
      if (0 == aid) return null;
      Assassin a;

      // check cache
      if (null == (a = (Assassin) assassinCache.get(new Integer(aid)))) {
         // check database existence
         // String query = "SELECT aid FROM " + DatabaseBackend.getPrefix() + "_assassins WHERE aid = '"+(new Integer(aid)).toString()+"'";
         // ResultSet r = DatabaseBackend.query(query);
         // try {
         //    if (!r.next()) throw new NonExistentEntityException();
         //    r.close();
         // } catch(SQLException SQLe) { throw new AssassinSQLException(SQLe); }

         // add to cache
         a = new Assassin(aid);
         assassinCache.put(new Integer(aid), a);
      }
      return a;
   }

   /** Called via reflection to get an instance of this by database ID */
   static Object getByID(Integer id) throws AssassinSQLException
   {
      return getByAid(id.intValue());
   }

   /**
    * Checks all Assassin objects for updates, and writes them out to the database.
    */
   static void commitAll() throws AssassinSQLException 
   {
      // iterate over all new Assassin objects
      while (newAssassinCache.size() > 0) {
         Assassin a = (Assassin) newAssassinCache.remove(0);
         a.writeDetails();
         assassinCache.put(new Integer(a.getAid()),a);
      }

      // iterate over all cached Assassin objects
      Iterator i = assassinCache.values().iterator();
      while (i.hasNext()) {
         Assassin a = (Assassin) i.next();
         // if they are modified, then update
         if (a.dirty()) a.writeDetails();
      }
   }
   
   /**
    * Removes an assassin from the database and the pending list.
    *
    * Be Careful with this! Do not reuse a reference if you have deleted it!!
    */
   static void deleteAssassin(Assassin a) throws AssassinSQLException
   {
      newAssassinCache.remove(a);
      assassinCache.remove(a);
      if (0 != a.id) {
         String query = "DELETE FROM "+DatabaseBackend.getPrefix()+"_assassins WHERE aid = '"+a.id+"'";
         try{
			 DatabaseBackend.query(query).close();
         } catch (SQLException sqle) {
			 throw new AssassinSQLException(sqle);
         }
         a.id = 0;
      }
   }


   private Player player = null;
    
   Assassin(int aid, Player player)
   {
      this.id = aid;
      this.player = player;
      assassinCache.put(new Integer(aid), this);
      clean();
   }

   Assassin(int aid)
   {
      this.id = aid;
      clean();
   }

   Assassin(Integer aid)
   {
      this(aid.intValue());
   }

   public Assassin()
   {
      this.id = 0;
      this.player = null;
      clean();
      newAssassinCache.add(this);
   }
   
   /**
    * Determine which player this assassin is played by.
    * @return The player object for the player who plays this assassin.
    */
   public Player getPlayer() throws AssassinSQLException
   {
      if (null == player) readDetails();
      if (Debug.debug) Debug.print(Assassin.class, ""+player);
      return this.player;
   }
   
   void setPlayer(Player p)
   {
      this.player = p;
      setDirty();
   }

   public int getAid() {
      return id;
   }

   public Viewer getViewer()
   {
      return new AssassinViewer(this);
   }
   
   void readDetails() throws AssassinSQLException
   {
      if (dirty()) throw new ObjectModifiedException();

      if (0 == id) throw new NoDatabaseAssociativityException();

      String query = "SELECT pid FROM " + DatabaseBackend.getPrefix() + "_assassins WHERE aid = '"+id+"'";
      AResultSet r = DatabaseBackend.query(query);

      // in this case it is valid for there not to be an association in the table - it means this
      // assassin is scheduled for deletion.
      try {
         if (! r.encapsulated.next()) return;
         this.player = Player.getByPid(r.encapsulated.getInt("pid"));
         // NOTE: this next line is a speculative fix
         // for the assassin deletion problem. It may
         // not be neccessary, but could prevent it
         // occurring - mjj29
         //
         // looks like this may have fixed it. Do not 
         // remove this line for optimisations without thinking about it.
         // possibly player.addAssassin(this)? This causes
         // readDetails();
         this.player.readDetails(Player.READASS);
         r.close();
      } catch(SQLException SQLe) { throw new AssassinSQLException(SQLe); }
   }

   void writeDetails() throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(Assassin.class, "a.writeDetails()");
      // cases:
      // - aid == 0, player == null: ignore
      // - aid == 0, player != null: add
      // - aid != 0, player == null: delete
      // - aid != 0, player != null: update
      
      if (!dirty()) return;
      clean();
      if (newAssassinCache.contains(this)) newAssassinCache.remove(this);

      if (0 == id && null == player) return;

      String query;
      
      if (0 == id && null != player) {
         player.writeDetails();
         query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_assassins (aid,pid) VALUES ('"+id+"','"+player.getPid()+"')";
         DatabaseBackend.query(query);

         // get auto-assigned aid
         query = "SELECT  LAST_INSERT_ID() FROM " + DatabaseBackend.getPrefix() + "_assassins";
         AResultSet r = DatabaseBackend.query(query);
         try {
            if (!r.encapsulated.next()) throw new AssassinSQLException("Oh Ek! That shouldn't happen?!");
            id = r.encapsulated.getInt(1);
            r.close();
         } catch(SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      }

      if (0 != id && null == player) {
         query = "DELETE FROM " + DatabaseBackend.getPrefix() + "_assassins WHERE aid='"+id+"'";
         try {
			 DatabaseBackend.query(query).close();
         } catch (SQLException sqle) {
			 throw new AssassinSQLException(sqle);
         }
         assassinCache.remove(new Integer(id));
         id = 0;
      }
      
      if (0 != id && null != player) {
         player.writeDetails();
         try{
			 query = "DELETE FROM " + DatabaseBackend.getPrefix() + "_assassins WHERE aid='"+id+"'";
			 DatabaseBackend.query(query).close();
			 query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_assassins (aid,pid) VALUES ('"+id+"','"+player.getPid()+"')";
			 DatabaseBackend.query(query).close();
         } catch (SQLException sqle) {
			 throw new AssassinSQLException(sqle);
         }
      }

   }

   public String getPrintable() throws AssassinSQLException
   {
      return getPlayer().getName()+" (associated assassin #"+id+")";
   }

   /**
    * Count the number of deaths we've suffered before time t.
    */
   public int countDeathsAtTime(long t) throws AssassinSQLException {
      String query = "SELECT COUNT(scid) FROM "+DatabaseBackend.getPrefix()+"_statechanges INNER JOIN "+DatabaseBackend.getPrefix()+"_events ON eid=event WHERE class='org.ucam.srcf.assassins.Death' AND assassin='"+id+"' AND time<"+t+" AND active='1'";
      AResultSet r = DatabaseBackend.query(query);
      int ans;
      try {
         r.encapsulated.next();
         ans = r.encapsulated.getInt(1);
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException(sqle);
      }
      return ans;
   }

   /**
    * Count the number of deaths we've suffered before time t.
    */
   public int countKillsAtTime(long t) throws AssassinSQLException {
      String query = "SELECT COUNT(scid) FROM "+DatabaseBackend.getPrefix()+"_statechanges INNER JOIN "+DatabaseBackend.getPrefix()+"_events ON eid=event WHERE class='org.ucam.srcf.assassins.Death' AND field='killed-by' AND value='__::org.ucam.srcf.assassins.Assassin("+id+")' AND time<"+t+" AND active='1'";
      AResultSet r = DatabaseBackend.query(query);
      int ans;
      try {
         r.encapsulated.next();
         ans = r.encapsulated.getInt(1);
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException(sqle);
      }
      return ans;
   }
   
   /**
    * Is the assassin alive or dead at the given time?
    * @return true if the assassin is alive or indeterminate, false if dead
    */
   public boolean isAliveAtTime(long t) throws AssassinSQLException {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT class FROM "+sct+" INNER JOIN "+evt+" ON ev.eid = sc.event WHERE (class = 'org.ucam.srcf.assassins.Resurrection' OR class = 'org.ucam.srcf.assassins.Death') AND ev.time <= '"+t+"' AND ev.active = 1 AND assassin= " + id + " ORDER BY ev.time DESC LIMIT 1";
      AResultSet r = DatabaseBackend.query(query);
      try {
         /* 
          * First check if there is one, otherwise indeterminate,
          * if you want to check you need to look for the 
          * AssassinActivation event for this assassin. We don't bother
          * here.
          */
         if(!r.encapsulated.next()) {
			 r.close();
			 return true;
         }
		 if("org.ucam.srcf.assassins.Death".equals(r.encapsulated.getString("class"))) {
			 r.close();
			 return false;
		 }
		 r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException(sqle);
      }
      return true;   // last event was a resurrection
   }

   double sumPluginFieldAtTime(String p, String f, long t) throws AssassinSQLException {
      String sql = "SELECT sum(value) FROM "+DatabaseBackend.getPrefix()+"_statechanges INNER JOIN "+DatabaseBackend.getPrefix()+"_events ON eid=event WHERE active='1' AND pluginName='"+p+"' AND field='"+f+"' AND assassin='"+id+"'";
      AResultSet r = DatabaseBackend.query(sql);
      try {
         if(!r.encapsulated.next()) {
            if (Debug.debug) Debug.print(this, this+".sumPluginFieldAtTime("+p+","+f+","+t+") = 0");
            return 0;
         }
         if (Debug.debug) Debug.print(this, this+".sumPluginFieldAtTime("+p+","+f+","+t+") = "+r.encapsulated.getDouble(1));
         return r.encapsulated.getDouble(1);
      } catch (SQLException sqle) {
         throw new AssassinSQLException(sqle);
      }
   }
}


