/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.DatabaseNotConnectedException;
import org.ucam.srcf.assassins.exceptions.UnknownDatabaseTypeException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * This class manages access to the database. 
 * It provides a method to send arbitrary queries to the database.
 *
 * @author Matthew Johnson <autoumpire@matthew.ath.cx>
 */
class DatabaseBackend
{
   /** The Connection to the database */
   private static Connection connection = null;
   private static String databasePrefix = null;
   
   /**
    * Initialiase the connection to the database
    * 
    * @throws AssassinSQLException Thrown if there is a problem with the SQL transaction
    */
   static void init() throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Initialising Database Backend");
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Getting config");
      String type = Config.getConfig("vanilla","dbtype");
      String host = Config.getConfig("vanilla","dbhost");
      String port = Config.getConfig("vanilla","dbport");
      String name = Config.getConfig("vanilla","dbname");
      String user = Config.getConfig("vanilla","dbuser");
      String pass = Config.getConfig("vanilla","dbpass");
      databasePrefix = Config.getConfig("vanilla", "game_prefix").replace('-','_');
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Connecting to database");
      Properties p = new Properties();
      p.setProperty("user", user);
      p.setProperty("password", pass);
      Driver d;
      try {
         if (type.equals("mysql")) 
            d = new com.mysql.jdbc.Driver();
         else
            throw new UnknownDatabaseTypeException();
         connection = d.connect("jdbc:"+type+"://"+host+":"+port+"/"+name, p);
			//Connector/J has autocommit feature. we turn this off.
			connection.setAutoCommit(false);
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
   }

   /**
    * Creates a PreparedStatement.
    */
   public static PreparedStatement getPreparedStatement(String sql) throws AssassinSQLException {
      try {
         return connection.prepareStatement(sql);
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
   }

   /**
    * Performs a query on the database.
    *
    * @param q The query to perform
    * @return A ResultSet object containing the result of the query
    * @throws AssassinSQLException Thrown if there is a problem with the SQL transaction
    */
   public static AResultSet query(String q) throws AssassinSQLException
   {
      long start = 0;
      if (Debug.debug) Debug.print(DatabaseBackend.class, q);
      if (Debug.debug) start = System.currentTimeMillis();
      if (null == connection) throw new DatabaseNotConnectedException();
      AResultSet r=null;
      try {
         Statement s = connection.createStatement();
         s.execute(q);
         connection.commit();
         r=new AResultSet(s.getResultSet(), s);
         /* s.close(); 		Removed for Connector/J compat. */
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      if (Debug.debug) Debug.print(DatabaseBackend.class, 
            " (executed in: "+(System.currentTimeMillis()-start)+"ms)");
      return r;
   }

   /**
    * Returns the prefix for all the table names
    *
    * @return The table name prefix
    */
   public static String getPrefix()
   {
      return databasePrefix;
   }

   /**
    * Create the tables for the game.
    * Should be called once at the start of the game, after editing the config file
    *
    * <b>WARNING! - Will delete your data! Make _REALLY_ sure you want to do this!!</b>
    */
   static void initTables() throws AssassinSQLException
   {
      String query;
      query = "DROP TABLE " + databasePrefix + "_session_other_data;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_session_data;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_active_sessions;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_eventparticipants;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_reports;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_pseudonyms;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_statechanges;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_events;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_assassins;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      query = "DROP TABLE " + databasePrefix + "_players;";
      try {
         query(query).close();
      } catch (Exception asqle) {}
      try {
         query = "CREATE TABLE " + databasePrefix + "_players(pid  INT AUTO_INCREMENT PRIMARY KEY, name VARCHAR(50), college VARCHAR(30), email VARCHAR(50), address TEXT, waterstatus TEXT, notes TEXT, password VARCHAR(50), INDEX(pid)) Type=MyISAM;";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_assassins(aid INT AUTO_INCREMENT PRIMARY KEY, pid INT, INDEX(pid), FOREIGN KEY(pid) REFERENCES "+databasePrefix+"_players(pid),INDEX(aid)) Type=MyISAM;";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_events(eid INT AUTO_INCREMENT PRIMARY KEY, headline VARCHAR(255), time BIGINT, modified BIGINT, active BOOL, visible BOOL) Type=MyISAM;";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_statechanges(scid INT AUTO_INCREMENT PRIMARY KEY, class VARCHAR(60), event INT, assassin INT, pluginName VARCHAR(20), field VARCHAR(20), processed BOOL, value VARCHAR(255), INDEX(event), INDEX(assassin), INDEX(field,pluginName), FOREIGN KEY(event) REFERENCES " + databasePrefix + "_events(eid), FOREIGN KEY(assassin) REFERENCES " + databasePrefix + "_assassins(aid)) Type=MyISAM;";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_pseudonyms(psid INT AUTO_INCREMENT PRIMARY KEY, assassin INT, pseudonym TEXT, INDEX(assassin), INDEX(psid), FOREIGN KEY(assassin) REFERENCES " + databasePrefix + "_assassins(aid)) Type=MyISAM;";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_reports(rid INT AUTO_INCREMENT PRIMARY KEY, event INT, assassin INT, pseudonym INT, INDEX(event), INDEX(assassin),INDEX(pseudonym), text TEXT, FOREIGN KEY(event) REFERENCES " + databasePrefix + "_events(eid), FOREIGN KEY(assassin) REFERENCES " + databasePrefix + "_assassins(aid), FOREIGN KEY(pseudonym) REFERENCES " + databasePrefix + "_pseudonyms(psid)) Type=MyISAM;";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_eventparticipants(aid INT NOT NULL, eid INT NOT NULL, psid INT, pindex INT, INDEX(eid), INDEX(aid), INDEX(psid), FOREIGN KEY(eid) REFERENCES " + databasePrefix + "_events(eid), FOREIGN KEY(aid) REFERENCES " + databasePrefix + "_assassins(aid), FOREIGN KEY(psid) REFERENCES " + databasePrefix + "_pseudonyms(psid), PRIMARY KEY(aid,eid)) Type=MyISAM;";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_active_sessions(sid INT AUTO_INCREMENT PRIMARY KEY, pid integer not null, index (pid), foreign key(pid) references " + databasePrefix + "_players (pid), session_id VARCHAR(50), expires DATETIME, auth_level INT);";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_session_data(did INT AUTO_INCREMENT PRIMARY KEY, sid integer not null, index (sid), foreign key(sid) references " + databasePrefix + "_active_sessions (sid), plugin VARCHAR(20), index(plugin), question TEXT);";
         query(query).close();
         query = "CREATE TABLE " + databasePrefix + "_session_other_data(sid INT NOT NULL, index(sid), foreign key(sid) references "+databasePrefix+"_active_sessions (sid), field VARCHAR(20), value TEXT)";
         query(query).close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException(sqle);
      }
   }

   static void commitAll() throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Commiting Pseudonyms");
      Pseudonym.commitAll();
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Commiting Players");
      Player.commitAll();
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Commiting Assassins");
      Assassin.commitAll();
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Commiting Events");
      Event.commitAll();
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Commiting Statechanges");
      StateChange.commitAll();
      if (Debug.debug) Debug.print(DatabaseBackend.class, "Commiting Reports");
      Report.commitAll();
   }

   static boolean validateUser(String email, String password)
   {
      if (null == email || null == password || "" == email || "" == password) return false;
      String query = "SELECT pid FROM "+databasePrefix+"_players WHERE email='"+Tools.escape(email)+"' AND password='"+Tools.escape(password)+"'";
      try {
         AResultSet rs = query(query);
         if (!rs.encapsulated.next()) {
            rs.close();
            return false;
         }
		   rs.close();
         return true;
      } catch (Exception e) {
         return false;
      }
   }

   static boolean updateAuthenticationDetails(String email, String password)
   {
      try {
         // make sure the user exists
         String query = "SELECT pid FROM "+databasePrefix+"_players WHERE email='"+email+"'";
         AResultSet r = query(query);
         if (! r.encapsulated.next()) return false;
         r.close();

         // set their password
         query = "UPDATE "+databasePrefix+"_players SET password='"+password+"' WHERE email='"+email+"'";
         r = query(query);
         r.close();
      } 
      // return false on error
      catch (AssassinSQLException ASQLe) { if (Debug.debug) Debug.print(DatabaseBackend.class, "AssassinSQLException: "+ASQLe); return false; }
      catch (SQLException SQLe) { if (Debug.debug) Debug.print(DatabaseBackend.class, "SQLException: "+SQLe); return false; }
      return true;
   }
}

