/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;

/**
 * Viewer for the event class.
 */
public class EventViewer extends CanonicalViewer
{
   EventViewer(Event e)
   {
      this.ent = e;
   }
   
   /**
    * Retrieve a mapping of participant indices to assassins for this event.
    * Each participant is assigned a unique index for the event so that it
    * can be used in [P1] tags and so on without having to worry about what
    * happens if other assassins are associated with or dissociated from
    * the event.
    * @return A map of participant indices to {@link AssassinViewer}s.
    */
   public Map getParticipants() throws AssassinSQLException
   {
      Map res = new TreeMap();
      Map src = ((Event)this.ent).getParticipants();
      Iterator i = src.keySet().iterator();
      while (i.hasNext())
      {
         Object idx = i.next();
         Assassin a = (Assassin) src.get(idx);
         res.put(idx, new AssassinViewer(a));
      }
      return res;
   }
   
   /**
    * Determine which assassin a given participant index refers to.
    * @param which The participant index to ask about.
    * @return A viewer for the assassin in question.
    */
   public AssassinViewer getParticipant(int which) throws AssassinSQLException
   {
      return new AssassinViewer(((Event)this.ent).getParticipant(which));
   }
   /**
    * Retrieve which pseudonym is being used by a particular participant.
    * @param which The participant index to ask about.
    * @return A string representing the pseudonym being used.
    */
   public String getParticipantPseudonym(int which) throws AssassinSQLException
   {
      return ((Event)this.ent).getParticipantPseudonym(which).getValue();
   }

	/**
	 * Tells you whether the event is active.
	 * @return true if it is, false if not
	 */
	public boolean isActive() throws AssassinSQLException 
	{
		return ((Event) this.ent).isActive();
	}

	/**
	 * Tells you whether the event is visible.
	 * @return true if it is, false if not
	 */
	public boolean isVisible() throws AssassinSQLException
	{
		return ((Event) this.ent).isVisible();
	}

   /**
    * Get a list of state changes attached to this event.
    * @return A list of state changes attached to the event.
    */
   public List getStateChanges() throws AssassinSQLException
   {
      return ((Event)this.ent).getStateChanges();
   }

   /**
    * Retrieve the time at which this event occurs.
    * @return The time at which the event occurs (in seconds since the epoch).
    */
   public long getTime() throws AssassinSQLException
   {
      return ((Event)this.ent).getEventTime();
   }

   /**
    * Retrieve the headline for this event.
    * @return The headline string for the event.
    */
   public String getHeadline() throws AssassinSQLException
   {
      return ((Event)this.ent).getHeadline();
   }
}

