/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See he
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.NoDatabaseAssociativityException;
import org.ucam.srcf.assassins.exceptions.ObjectModifiedException;

/**
 * Class to represent a pseudonym. I think this ends up being less icky than
 * assigning pseudonym indices on a per-assassin basis because every pseudonym
 * can be identified uniquely.
 *
 * At the moment a pseudonym (as presented to the user in Java) makes no
 * attempt to keep track of how many or which assassins own it. I don't think
 * there's ever a situation where this is necessary, and it means that the
 * Pseudonym class can be entirely canonical.
 *
 * @author Matthew Johnson <autoumpire@matthew.ath.cx>
 */
public class Pseudonym extends CanonicalEntity
{
   private static Map pseudonymCache = new HashMap();
   private static Vector newPseudonymCache = new Vector();
   private static boolean preCacheState = false;
   
   /**
    * Loads all the objects into memory frm the database in one query
    */
   static void preCache() throws AssassinSQLException
   {
      if (preCacheState) return;
      preCacheState = true;
      AResultSet r = DatabaseBackend.query("SELECT psid, assassin, pseudonym FROM "+DatabaseBackend.getPrefix()+"_pseudonyms");
      try {
         while (r.encapsulated.next()) {

            int psid = r.encapsulated.getInt("psid");
            Integer PSID = new Integer(psid);

            // check the cache and add as necessary
            if (!pseudonymCache.containsKey(PSID)) {
               Pseudonym p = new Pseudonym(psid);
               p.assassin = Assassin.getByAid(r.encapsulated.getInt("assassin"));
               p.pseud = r.encapsulated.getString("pseudonym");
               pseudonymCache.put(PSID, p);
            }
         }
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException();
      }
   }
    
   /**
    * Returns all the pseudonyms
    */
   static List getAllPseudonyms() throws AssassinSQLException
   {
      preCacheState = true;
      List ps = new Vector();
      AResultSet r = DatabaseBackend.query("SELECT psid, assassin, pseudonym FROM "+DatabaseBackend.getPrefix()+"_pseudonyms");
      try {
         while (r.encapsulated.next()) {

            int psid = r.encapsulated.getInt("psid");
            Integer PSID = new Integer(psid);

            // check the cache and add as necessary
            if (pseudonymCache.containsKey(PSID)) 
               ps.add(pseudonymCache.get(PSID));
            else {
               Pseudonym p = new Pseudonym(psid);
               p.assassin = Assassin.getByAid(r.encapsulated.getInt("assassin"));
               p.pseud = r.encapsulated.getString("pseudonym");
               pseudonymCache.put(PSID, p);
               ps.add(p);
            }
         }
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException();
      }
      return ps;
   }
   

   /**
    * Returns a Pseudonym with a specific database ID
    *
    * @param psid The Database ID
    * @return The Pseudonym object
    */
   static Pseudonym getByPsid(int psid) throws AssassinSQLException
   {
      if (0 == psid) return null;
      Pseudonym p;

      // check cache
      if (null == (p = (Pseudonym) pseudonymCache.get(new Integer(psid)))) {
         // check database existence
         // String query = "SELECT psid FROM " + DatabaseBackend.getPrefix() + "_pseudonyms WHERE psid = '"+psid+"'";
         // ResultSet r;
         // try {
         //    r = DatabaseBackend.query(query);
         //    if (!r.next()) throw new NonExistentEntityException("Eek");
         //    r.close();
         // } catch (SQLException sqle) {
         //    throw new AssassinSQLException(sqle);
         // }

         // add to cache
         p = new Pseudonym(psid);
         pseudonymCache.put(new Integer(psid), p);
      }
      return p;

   }
   
   /**
    * Returns a Pseudonym with a specific database ID
    *
    * @param id The Database ID
    * @return The Pseudonym object
    */
   static Object getByID(Integer id) throws AssassinSQLException
   {
      return getByPsid(id.intValue());
   }
  
   /**
    * Commit all changes to Pseudonyms to the database
    */
   static void commitAll() throws AssassinSQLException
   {
      // iterate over all new Pseudonym objects
      while (newPseudonymCache.size() > 0) {
         Pseudonym a = (Pseudonym) newPseudonymCache.remove(0);
         a.writeDetails();
         pseudonymCache.put(new Integer(a.getPsid()),a);
      }
     
      // iterate over all cached Pseudonym objects
      Iterator i = pseudonymCache.values().iterator();
      while (i.hasNext()) {
         Pseudonym p = (Pseudonym) i.next();
         // if they are modified, then update
         if (p.dirty()) p.writeDetails();
      }
   }
   
   /**
    * Removes a pseudonym from the database and the pending list.
    *
    * Be Careful with this! Do not reuse a reference if you have deleted it!!
    */
   static void deletePseudonym(Pseudonym p) throws AssassinSQLException
   {
      newPseudonymCache.remove(p);
      pseudonymCache.remove(p);
      if (0 != p.id) {
         String query = "DELETE FROM "+DatabaseBackend.getPrefix()+"_pseudonyms WHERE psid = '"+p.id+"'";
         try {
			 DatabaseBackend.query(query).close();
         } catch (SQLException sqle) {
			 throw new AssassinSQLException(sqle);
         }
         p.id = 0;
      }
   }

   private String pseud = null;
   private Assassin assassin = null;
  
   public Pseudonym()
   {
      this.id = 0;
      this.pseud = "";
      newPseudonymCache.add(this);
      clean();
   }
   
   Pseudonym(Integer psid)
   {
      this.id = psid.intValue();
      clean();
   }

   Pseudonym(int psid)
   {
      this.id = psid;
      clean();
   }

   /**
    * Return the assassinwho owns this pseudonym.
    * @return an Assassin
    */
   public Assassin getAssassin() throws ObjectModifiedException, NoDatabaseAssociativityException, AssassinSQLException
   {
      try {
         if (null == assassin) {
            String query = "select assassin from "+DatabaseBackend.getPrefix()+"_statechanges s inner join "+DatabaseBackend.getPrefix()+"_events e on e.eid = s.event where field='pseudonym-gain' and pluginName='vanilla' and value='"+toString()+"' order by time desc limit 1;";
            AResultSet r = DatabaseBackend.query(query);
            if (!r.encapsulated.next()) return null;
            assassin = Assassin.getByAid(r.encapsulated.getInt("assassin"));
            r.close();
         }
         return assassin;
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
   }
 
   /**
    * Return this pseudonym as a string.
    * @return The pseudonym.
    */
   public String getValue() throws ObjectModifiedException, NoDatabaseAssociativityException, AssassinSQLException
   {
      if (null == pseud) readDetails();
      return pseud;
   }
  
   public int getPsid() {
      return id;
   }
    
   /**
    * set the assassin associated with this pseudonym.
    * @param a The new Assassin.
    */
   void setAssassin(Assassin a)
   {
      setDirty();
      assassin = a;
   }


   /**
    * Change this pseudonym.
    * @param value The new text of the pseudonym.
    */
   void setValue(String value)
   {
      setDirty();
      pseud = value;
   }

   /**
    * Update the details of this from the database
    */
   void readDetails() throws ObjectModifiedException, NoDatabaseAssociativityException, AssassinSQLException
   {
      if (dirty()) throw new ObjectModifiedException();

      if (0 == id) throw new NoDatabaseAssociativityException();

      String query = "SELECT assassin, pseudonym FROM " + DatabaseBackend.getPrefix() + "_pseudonyms WHERE psid = '"+id+"'";
      AResultSet r;
      
      try {
         r = DatabaseBackend.query(query);
         if (! r.encapsulated.next()) throw new NoDatabaseAssociativityException();
         this.pseud = r.encapsulated.getString("pseudonym");
         this.assassin = Assassin.getByAid(r.encapsulated.getInt("Assassin"));
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException (sqle);
      }
   }
  
   /**
    * Update the details of this int the database
    */
   void writeDetails() throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(Pseudonym.class, "Writing Pseudonym");
      if (!dirty()) return;
      clean();
      if (newPseudonymCache.contains(this)) newPseudonymCache.remove(this);

      // insert new
      if (0 == id) {
         String query;
         if (null == assassin)
            query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_pseudonyms (pseudonym) VALUES ('"+Tools.escape(pseud)+"')";
         else {
            assassin.writeDetails();
            query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_pseudonyms (pseudonym, assassin) VALUES ('"+Tools.escape(pseud)+"', '"+assassin.getAid()+"')";
         }
         DatabaseBackend.query(query);
         query = "SELECT LAST_INSERT_ID() FROM " + DatabaseBackend.getPrefix() + "_pseudonyms";
         AResultSet r;
         try {
            r = DatabaseBackend.query(query);
            if (!r.encapsulated.next()) throw new AssassinSQLException("Oh Ek! That shouldn't happen?!");
            id = r.encapsulated.getInt(1);
            r.close();
         } catch (SQLException sqle) {
            throw new AssassinSQLException(sqle);
         }

      // update existing
      } else {
         if (null == pseud) return;
         String query;
         if (null == assassin)
            query = "UPDATE " + DatabaseBackend.getPrefix() + "_pseudonyms SET pseudonym = '"+Tools.escape(pseud)+"' WHERE psid = '"+id+"'";
         else {
            assassin.writeDetails();
            query = "UPDATE " + DatabaseBackend.getPrefix() + "_pseudonyms SET pseudonym = '"+Tools.escape(pseud)+"', assassin = '"+assassin.getAid()+"' WHERE psid = '"+id+"'";
         }
         DatabaseBackend.query(query);
      }

   }
   
   public Viewer getViewer()
   {
      return null;
   }
   
   public String getPrintable() throws AssassinSQLException
   {
      return getValue();
   }

}

