package org.ucam.srcf.assassins.coffee;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.ucam.srcf.assassins.AssassinViewer;

/**
 * Class implementing a tree of kills on a per Assassin Basis.
 * Should handle circular kills in a good fashion.(tm).
 */
public class KillTree {
	/* Implementation notes:
	 *		We use recursive ArrayLists to form the map, these are
	 *		embedded in the AssassinVictims class.
	 *		however for random access adding we keep a HashMap
	 *		from AssassinViewers to their lists. We use null to represent
	 *		the killer of someone who hasn't died, or was killed with no 
	 *		killer (eg thunderbolted by the umpire).
	 *		We have to be careful of cases where we can get loops
	 *		and also remember people can still kill others even after 
	 *		they themselves are dead.
	 *		We deal with this by after we have built our tree we assign
	 *		every assassin a unique number and then take min's up the
	 *		tree [having assigned from the top down in increasing order]
	 *		and keep going until no more changes. Any assassins with the
	 *		same number are part of the same kill loop.
	 *		i.e. if A killed B then A's vlaue is min(A_val, B_val),
	 *		however if B also killed A then his value is min(B_val, A_val)
	 *		and we see they will both be equal.
	 *		If we detect this in the scoring we divide give everyone in
	 *		the group n+m kills, where m is the number beneath the group
	 *		and n is the number in the group.
	 */
	/**
	 * Class to represent a list of victims.
	 * Also who killed them and their uniqueness.
	 */
	private class AssassinVictims {
		public int uniqueness_val=0;
		private List victims=new ArrayList();
		private AssassinViewer who;
		public AssassinVictims killer=null;
		public int seenval=0;
		public boolean part_of_cycle=false;;
		public int score;

		AssassinVictims(AssassinViewer who) {
			this.who=who;
		}
		public AssassinViewer getWho() {
			return who;
		}
		public void addVictim(AssassinVictims body) {
			victims.add(body);
		}
		public List getVictims() {
			return victims;
		}
		public void removeVictim(AssassinVictims not_mine) {
			victims.remove(not_mine);
		}
	}
	private AssassinVictims topLevel=new AssassinVictims(null);;
	private Map assassins = new HashMap();
	private int nextUVal;
	private int currentSeenVal=1;

	/**
	 * Adds a death to the tree.
	 */
	public void addDeath(AssassinViewer killer, AssassinViewer victim) {
		AssassinVictims killersVictims = getVictims(killer);
		AssassinVictims victimsVictims = getVictims(victim);
		if(victimsVictims.killer==null)
			topLevel.removeVictim(victimsVictims);
		else
			victimsVictims.killer.removeVictim(victimsVictims);
		victimsVictims.killer=killersVictims;
		killersVictims.addVictim(victimsVictims);
	}

	private AssassinVictims getVictims(AssassinViewer killer) {
		AssassinVictims theList = (AssassinVictims) assassins.get(killer);
		if(theList==null)
			topLevel.addVictim(theList=new AssassinVictims(killer));
		return theList;
	}

	private List getCycles() {
		/*
		 * Start at the top and number "down", stopping if this 
		 * ones already been numbered. Then min down the tree until 
		 * nothing more changes, then return a list of cycles.
		 * A cycle being represented by a list of thsoe involved.
		 */

		/*zero them*/
		Iterator i=assassins.keySet().iterator();
		while(i.hasNext())
			((AssassinVictims) i.next()).uniqueness_val=0;
		
		/* number them all */
		nextUVal=1;
		numberTree(topLevel);
		
		/* min them all the way down, get a true back if there was a change
		 * and if so go again */
		while(minTree(topLevel, false));

		/* now parse and return a list of cycles, we build a map from values
		 * to lists of ones having that value */
		i=assassins.keySet().iterator();
		Map values=new HashMap();
		while(i.hasNext()) {
			AssassinVictims current = (AssassinVictims) i.next();
			List val_list = (List) values.get(new Integer(current.uniqueness_val));
			if(val_list == null)
				values.put(new Integer(current.uniqueness_val), val_list=new ArrayList());
			val_list.add(current);
		}

		/*and now go through all the values and return any non 1 length one ones.*/
		i=values.keySet().iterator();
		List ret_val=new ArrayList();
		while(i.hasNext()) {
			List current = (List) i.next();
			if(current.size() > 1)
				ret_val.add(current);
		}

		return ret_val;
	}

	private void numberTree(AssassinVictims to_n) {
		if(to_n.uniqueness_val!=0)
			return;	/*already done*/
		to_n.uniqueness_val=nextUVal++;
		Iterator i=to_n.getVictims().iterator();
		while(i.hasNext())
			numberTree((AssassinVictims) i.next());
	}

	private boolean minTree(AssassinVictims to_min, boolean changed) {
		/* we minimise this over all its victims and its own value
		 * and return true if it changes. The seen vlaue just tracks
		 * whether we've seen this on this iteration or not so we don't
		 * end up going in circles*/
		if(to_min.seenval>=currentSeenVal)
			return changed;
		to_min.seenval=currentSeenVal;
		Iterator i=to_min.getVictims().iterator();
		int min=to_min.uniqueness_val;
		boolean new_c=changed;
		while(i.hasNext()) {
			AssassinVictims current = (AssassinVictims) i.next();
			new_c = minTree(current, changed);
			if(current.uniqueness_val < min)
				min=current.uniqueness_val;
		}
		if(min < to_min.uniqueness_val) {
			to_min.uniqueness_val=min;
			new_c=true;
		}
		return new_c;
	}

	/**
	 * Calulates score in conkers terms.
	 * ie number of people underneath any given preson in the tree.
	 * @return Map from AssassinViewer to score
	 */
	public Map scoreAll() {
		/*first we count for everyone not on the list of cycles*/
		/*mark them as in a cycle if they are*/
		List cycles = getCycles();
		Iterator i = cycles.iterator();
		while(i.hasNext()) {
			Iterator j = ((List) i.next()).iterator();
			while(j.hasNext())
				((AssassinVictims) j.next()).part_of_cycle=true;
		}

		/*We now recurse thourgh and score em all*/
		i=topLevel.getVictims().iterator();
		while(i.hasNext())
			scoreTree((AssassinVictims) i.next(), cycles);

		/*Finally build and return the map*/
		i=assassins.keySet().iterator();
		Map ret_vals=new HashMap();
		while(i.hasNext()) {
			AssassinVictims current = (AssassinVictims) i.next();
			AssassinViewer currentAV = current.getWho();
			ret_vals.put(currentAV, new Integer(current.score));
		}

		return ret_vals;
	}
	
	private void scoreTree(AssassinVictims what, List cycles) {
		/*If its a cycle score everything below it first, then score 
		 * them all at once, otherwise just call self recursivly */
		if(what.part_of_cycle) {
			int cycle_total=0;
			/*find self in list*/
			Iterator i=cycles.iterator();
			List current=null;
			while(i.hasNext()) {
				current = (List) i.next();
				if(current.contains(what))
					break;
			}
			i=current.iterator();
			while(i.hasNext()) {		
				/*score everything with differnt uniqueness value we've 
				 * killed and add it to the cycle's total*/
				AssassinVictims currentMember = (AssassinVictims) i.next();
				Iterator j = currentMember.getVictims().iterator();
				while(j.hasNext()) {
					AssassinVictims currentVictim = (AssassinVictims) j.next();
					if(currentVictim.uniqueness_val == 
							currentMember.uniqueness_val)
						continue;
					scoreTree(currentVictim, cycles);
					cycle_total += currentVictim.score + 1;
				}
			}
			/*Now set everyone in the cycle's totla to n+m*/
			cycle_total+=current.size();
			i=current.iterator();
			while(i.hasNext())
				((AssassinVictims) i.next()).score=cycle_total;
		}
		else {
			Iterator i=what.getVictims().iterator();
			int score=0;
			while(i.hasNext()) {
				AssassinVictims currentVictim = (AssassinVictims) i.next();
				scoreTree(currentVictim, cycles);
				score += currentVictim.score+1;
			}
			what.score = score;
		}
	}
}
