/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.mulberry;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.ConfigProperty;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.OptionQuestion;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginCommand;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.YesNoQuestion;
import org.ucam.srcf.assassins.autoumpire;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * Mafia game plugin.
 */
public class Mulberry extends Plugin
{
   public String getDescription() { return "Mafia game plugin. Assigns assassins to different classes."; };
   private List classes;
   private List numbers;
   private boolean autoassign;

   public void setPluginName(String s) 
   {
      super.setPluginName(s);
      String a = Config.getConfig(getPluginName(), "auto-assign");      
      autoassign = false;
      if (null != a) {
         if (a.toUpperCase().equals("TRUE")) autoassign = true;
         if (a.toUpperCase().equals("YES")) autoassign = true;
      }
      int i = 1;
      classes = new LinkedList();
      numbers = new LinkedList();
      String name = Config.getConfig(getPluginName(), "class-"+i+"-name");
      String number = Config.getConfig(getPluginName(), "class-"+i+"-number");
      while (name != null) { // might be ""
         if (Debug.debug) Debug.print(Mulberry.class, "Class "+i+": "+name+" ("+number+")");
         classes.add(name);
         if (null != number) try {
            numbers.add(new Integer(number));
         } catch(NumberFormatException NFe) {
            numbers.add(new Integer(0));
         }
         i++;
         name = Config.getConfig(getPluginName(), "class-"+i+"-name");
         number = Config.getConfig(getPluginName(), "class-"+i+"-number");
      }
   }
   
   /**
    * Called whenever the plugin is required to update or sanity-check the
    * current gamestate.
    */
   protected Vector updateState(Map state, Vector qs, EventViewer e) throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(Mulberry.class, "Updating Mulberry State");
      Vector nq = new Vector();
      
      if (qs.size() == 0) {
         // first, check everyone and either ask questions or calculate auto-assign
         
         for (Iterator i = state.values().iterator(); i.hasNext(); ) {
            AssassinStateViewer asv = (AssassinStateViewer) i.next();
            if (null == asv.getProperty(this, "class")) {
               if (autoassign) {
                  if (Debug.debug) Debug.print(Mulberry.class, "Auto-Assigning for "+asv.getAssassin().getPlayerViewer().getName());
                  // TODO: randomly assign based on weighting.
               } else {
                  nq.add(new OptionQuestion("Select Class for "+asv.getAssassin().getPlayerViewer().getName(),
                                            classes, 0));
               }
            }
         }
                         
         if (nq.size() == 0) return null;
         return nq;
      }

         // second, get the answers to questions and apply them
         Map classmap = new HashMap();
         for (Iterator i = qs.iterator(); i.hasNext(); ) {
            OptionQuestion q = (OptionQuestion) i.next();
            classmap.put(q.getLabel(), q.getAnswerString());
         }
         
         for (Iterator i = state.values().iterator(); i.hasNext(); ) {
            AssassinStateViewer asv = (AssassinStateViewer) i.next();
            if (null == asv.getProperty(this, "class")) {
               String cls = (String) classmap.get("Select Class for"+asv.getAssassin().getPlayerViewer().getName());
               if (Debug.debug) Debug.print(Mulberry.class,
                     "Setting class to "+cls+" for "+asv.getAssassin().getPlayerViewer().getName());
               if (null != cls)
                  vanilla.addStateChange(e, new PluginStateChange(asv.getAssassin(), this, "class", cls));
            }
         }
         return null;
   }
   
   /**
    * Called by the filtering mechanism when the plugin needs to determine
    * whether an assassin has a given property such as &quot;police&quot;. If
    * the plugin does not recognise the named property, it should return
    * <code>false</code>.
    * @param a A viewer for the assassin being tested.
    * @param prop The name of the property being asked about.
    * @return True if the assassin has the property, false if not.
    */
   public boolean assassinHasProperty(AssassinStateViewer asv, String prop, String value)
   {
      if (Debug.debug) Debug.print(autoumpire.class, "assassinHasProperty("+asv+","+prop+","+value+")");
      if (!"class".equals(prop)) return false;
      String cls = asv.getProperty(this, "class");
      if (null == cls) return false;
      if (cls.equals(value)) return true;
      return false;
   }

   /**
    * Called when the system has decided that email needs to be sent to a given
    * player. The plugin may return a string to be included in the email, or
    * null if it does not wish to include anything.
    * @param p A viewer for the player being emailed.
    * @return A string to be included in the email sent to the player, or null.
    *    The string will appear in its own paragraph when inserted into the
    *    email and so should not begin or end with a newline.
    */
   protected String generateEmail(AssassinStateViewer asv) 
   {
      return "";
   }

   /**
    * Called when an assassin is being added to get any queries the plugin
    * has for the user. The plugin should check all the assassin details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param a The Assassin which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector assassinAdded(AssassinViewer a, EventViewer e, Map state, Vector qs) throws AssassinSQLException 
   {
      if (null == qs) {
         qs = new Vector();
         qs.add(new YesNoQuestion("Give Class",false));
         qs.add(new OptionQuestion("Class",classes,0));
         return qs;
      } else {

         boolean do_cls = ((YesNoQuestion)qs.get(0)).getAnswer();   
         if (do_cls) {
            String cls = ((OptionQuestion)qs.get(1)).getAnswerString();
            vanilla.addStateChange(e, new PluginStateChange(a, this, "class", cls));
         }
         return null;
      }
   }

   // CLI method.
   protected void changeClass(Map params) throws AssassinSQLException
   {
      // TODO: add a change class interface
   }

   /**
    * Returns a Map of new commands to add to the CLI
    *
    * @return A Map of String =&gt; Command
    */
   public Map getCommands()
   { 
      Map m = new HashMap();
      final Mulberry mul = this;
      m.put("change-class",
             new PluginCommand () {
               public void pluginCommand(Map params) throws AssassinSQLException
               {
                  mul.changeClass(params);
               }  
            }
            );
      return m;
   }
   
   public  static Vector getConfigProperties() 
   { 
      Vector v = new Vector(); 
      v.add ( new ConfigProperty ("auto-assign", "Automatically Assign Classes", "true"));
      return v;
   }


}


