/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.strawberry;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ucam.srcf.assassins.AssassinActivation;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.Event;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.PluginConnector;
import org.ucam.srcf.assassins.Resurrection;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.Tools;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.ConfigFileException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class StrawberryDynamic
{
	Scorer AbaseScore;
	Scorer VbaseScore;
	Scorer AScoreMod;
	Scorer VScoreMod;

   Strawberry strawberry;
   PluginConnector vanilla;

   void setPluginConnector(PluginConnector p) {
      vanilla=p;
   }

   StrawberryDynamic (Strawberry st, PluginConnector vanilla) throws ConfigFileException {
      strawberry = st;
      AbaseScore=new Scorer(Config.getConfig(strawberry.getPluginName(), "AbaseScoreEqn"),0);
      VbaseScore=new Scorer(Config.getConfig(strawberry.getPluginName(), "VbaseScoreEqn"),0);
      AScoreMod=new Scorer(Config.getConfig(strawberry.getPluginName(), "AScoreModEqn"),0);
      VScoreMod=new Scorer(Config.getConfig(strawberry.getPluginName(), "VScoreModEqn"),0);
      this.vanilla=vanilla;
   }
	
   /**
    * Scores the game from a given time, only inspecting assassins in a given
    * set.
    */
   void scoreFromSetFromTime(Set interested, long fromTime) throws AssassinSQLException {
		/* We proceed to go through the system and check the scores 
		 * assigned at each "death" event. This officially sucks, as we have
		 * to build our own state for aliveness
		 */
		/* I have been assured that noone would be silly enough to think
		 * that there should eb intermediate scores worked out during an
		 * event, and that noone will put a chain of deaths
		 * A->B->C and expect A to get some of B's points for killing
		 * C from this event.
		 */

		/* Set up a var tracker to track stuff */
		VarTracker vt = new VarTracker(interested, strawberry.getPluginName());

		/* Get all events in chronological order */
		List events = Event.getEventViewersInTime(fromTime, vanilla.getNowTime());

      /* Read in starting scores if we're not at the beginning */
      if(strawberry.startTime != fromTime) {
         if(Debug.debug) Debug.print(this, "Reading in starting scores for assassins");
         Iterator i = interested.iterator();
         while(i.hasNext()) {
            vt.readValuesForAssassinAtTime(strawberry, (AssassinViewer) i.next(), fromTime);
         }
      }
      
      Event.preCacheSubClasses(Event.READALL);
      StateChange.preCache();
		Iterator i = events.iterator();
		if(Debug.debug) Debug.print(this, "# events in time ("+fromTime+","+vanilla.getNowTime()+") = "+events.size());
		while(i.hasNext()) {
			EventViewer e = (EventViewer) i.next();
			if(!e.isActive())
				continue;
			if(Debug.debug) Debug.print(this, "Processing event ("+e+") \""+e.getHeadline()+"\" at time "+Tools.prettyDateAndTime(e.getTime()));
         if(!StrawberryTools.eventContainsParticipantsFromSet(e, interested)) {
            if(Debug.debug) Debug.print(this, "Doesn't contain anything we're interested in!");
            continue;
         }
			List sclist = e.getStateChanges();
			Iterator j = sclist.iterator();

         /* Flag for assassinActivation events */
         boolean assAct = false;

			/*Set up a map from assassins to *modified* scores*/
			Map scores = vt.refreshScores(e);
			
			/* List of score StateChanges so we can check if they need updating at 
			 * the end of processing an event
			 */
			List scoreChanges = new ArrayList();

			/*Now iterate through the statechanges*/
			while(j.hasNext()) {
				StateChange sc = (StateChange) j.next();
            if(!StrawberryTools.doesSCInvolveSet(sc, interested)) {
               if(Debug.debug) Debug.print(this, "We don't care about "+
                     "this one: " + sc);
               continue;
            }
				AssassinViewer a = sc.getAssassinViewer();
				String pname = sc.getPluginName();
				String field = sc.getFieldName();
				Object value = sc.getValue();
            if(sc instanceof AssassinActivation) {
               assAct = true;
               continue;
            }
				if(sc instanceof Death) {
					if(Debug.debug) Debug.print(this, "SC is a death");
					/* Make certain not already dead, assume alive if return is ""*/
               if(!vt.trackingAssassin(a))
                  vt.readValuesForAssassinAtTime(strawberry, a, sc.getEventViewer().getTime()-1);
					if(vt.getProperty(a,"alive").equals("false")) {
                  if(Debug.debug) Debug.print(this, "...but vicitim is already dead");
						continue;
               }
					vt.setProperty(a,"alive","false");
               /* Check if people were in the interested set and add 
                * them as appropriate, loading info */
               if(!interested.contains(a)) {
                  interested.add(a);
                  vt.readValuesForAssassinAtTime(strawberry, a, e.getTime());
               }
               if(!interested.contains(((Death) sc).getKiller())) {
                  interested.add((((Death) sc).getKiller()));
                  vt.readValuesForAssassinAtTime(strawberry, ((Death) sc).getKiller(), 
                        e.getTime());
               }
					/* Score it and notify the VarTracker. */
					scoreDeath((Death) sc, vt, scores);
					vt.trackDeath((Death) sc);
					continue;
				} 
				if(sc instanceof Resurrection) {
					if(Debug.debug) Debug.print(this, "SC is a resurrection");
					vt.setProperty(a,"alive","true");
					continue;
				} 
				if(!pname.equals(strawberry.getPluginName()))
					continue;
				/* Otherwise we have a sc applicable to this plugin, */
				if(Debug.debug) Debug.print(this, "SC is something else Strawberry related: "+field);

				/* If its a score change we deal with it later */
				if(	field.equals("BaseScore") ||
						field.equals("PostGameMod") ||
						field.equals("FinalScore")
						) {
					scoreChanges.add(sc);
					continue;
				}

				/* If its an innocent death we deal with it */
				if(field.equals("Innoccents")) {
					int numKilled = Integer.parseInt((String) value);
					for(int innocent=0; innocent<numKilled; innocent++) {
						scoreInnocentDeath(a, vt, scores);
						vt.trackInnocent(a);
					}
				}

				/*If its a bonus we add it to the scores*/
				if(field.equals("BaseBonus")) {
					if(Debug.debug) Debug.print(this, "Applying BaseBonus ...");
					AssassinScore as=((AssassinScore) scores.get(a));
					if(as==null) 
						scores.put(a, as=new AssassinScore(a, strawberry.getPluginName(), vt));
					as.addBase(Integer.parseInt((String) value));
					if(Debug.debug) Debug.print(this, "...done");
					continue;
				}
				if(field.equals("ModBonus")) {
					if(Debug.debug) Debug.print(this, "Applying ModBonus ...");
					AssassinScore as=((AssassinScore) scores.get(a));
					if(as==null)
						scores.put(a, as=new AssassinScore(a, strawberry.getPluginName(), vt));
					as.addPost(Integer.parseInt((String) value));
					if(Debug.debug) Debug.print(this, "...done");
					continue;
				}
				
				/* otherwise we just apply it */
				vt.setProperty(a, field, (String) value);
				if(Debug.debug) Debug.print(this, "Applying unknown Strawberry State Change (" + field + ", " + value + ")");
			} /*End of states loop */
			if(Debug.debug) Debug.print(this, "Done with states loop.");

			/* 
          * Now we go through and check if the scores are right or not
          * by going first going through preexisting ones and chekcing 
          * them, removing them if we have no record of them!
          */
			j = scoreChanges.iterator();
			while(j.hasNext()) {
				StateChange sc = (StateChange) j.next();
				AssassinViewer a = sc.getAssassinViewer();
				String field = sc.getFieldName();
				int value = Integer.parseInt((String) sc.getValue());
				if(scores.get(a) == null) {
               if(!assAct) {
                  /* Its wrong! remove it! */
                  vanilla.removeStateChange(e, sc);
               } else {
                  /* Its part of an assassin activation event */
                  vt.setProperty(a, field, String.valueOf(value));
               }
					continue;
				}
				AssassinScore cur = (AssassinScore) scores.get(a);
				if(field.equals("BaseScore")) {
					if(cur.getBaseScore() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.baseClean();
					continue;
				}
				if(field.equals("PostGameMod")) {
					if(cur.getPostGameMod() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.postClean();
					continue;
				}
				if(field.equals("FinalScore")) {
					if(cur.getFinalScore() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.finalClean();
					continue;
				}
			}
			if(Debug.debug) Debug.print(this, "Done with initial score storing loop.");
			/* Now we add any dirty score changes to the event */
			j = scores.keySet().iterator();
			while(j.hasNext()) {
				AssassinScore cur=(AssassinScore) scores.get(j.next());
				if(cur==null)
					continue;
				AssassinViewer a = cur.getAssassinViewer();
				/*write out if they need it*/
				if(cur.isBaseDirty())
					vanilla.addStateChange(e, cur.getBaseSC(strawberry));
				if(cur.isPostDirty())
					vanilla.addStateChange(e, cur.getPostSC(strawberry));
				if(cur.isFinalDirty())
					vanilla.addStateChange(e, cur.getFinalSC(strawberry));
				/*and update our internal state*/
				vt.setIntProperty(a, "BaseScore", cur.getBaseScore());
				vt.setIntProperty(a, "PostGameMod", cur.getPostGameMod());
				vt.setIntProperty(a, "FinalScore", cur.getFinalScore());
			}
			if(Debug.debug) Debug.print(this, "Done with final score storing loop.");
				
		} /*end of events loop */
	}

	private void scoreInnocentDeath(AssassinViewer a, VarTracker vt, Map scores) {
		int newAbaseScore = (int) Math.round(AbaseScore.execute(a, null, vt));
		int newApostGameMod = (int) Math.round(AScoreMod.execute(a, null, vt));
		
		AssassinScore aScore = (AssassinScore) scores.get(a);

		if(aScore==null)
			scores.put(a, aScore=new AssassinScore(a, strawberry.getPluginName(), vt));
		
		aScore.setBaseScore(newAbaseScore);
		aScore.setPostGameMod(newApostGameMod);
	}
	
	private void scoreDeath(Death myDeath, VarTracker vt, Map scores) {
      /* calculate the new scores */
      try {
			if(Debug.debug) Debug.print(this, "scoring death: "+myDeath.getKiller().getPlayerViewer().getName()+"->"+myDeath.getVictim().getPlayerViewer().getName());
         int newAbaseScore = (int) Math.round(AbaseScore.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newVbaseScore = (int) Math.round(VbaseScore.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newApostGameMod = (int) Math.round(AScoreMod.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newVpostGameMod = (int) Math.round(VScoreMod.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));

			if(Debug.debug) {
				Debug.print(this, "\tAbaseScore="+newAbaseScore);
				Debug.print(this, "\tApostGameMod="+newApostGameMod);
				Debug.print(this, "\tVbaseScore="+newVbaseScore);
				Debug.print(this, "\tApostGameMod="+newVpostGameMod);
			}

         /* get the current scores if they exist, otherwise make them */
         AssassinScore aScore = (AssassinScore) scores.get(myDeath.getKiller());
         AssassinScore vScore = (AssassinScore) scores.get(myDeath.getVictim());
         if (aScore == null)
               scores.put(myDeath.getKiller(), aScore = new AssassinScore(
                     myDeath.getKiller(), strawberry.getPluginName(), vt));
         if (vScore == null)
               scores.put(myDeath.getVictim(), vScore = new AssassinScore(
                     myDeath.getVictim(), strawberry.getPluginName(), vt));
         /* now set the scores correctly */
         aScore.setBaseScore(newAbaseScore);
         aScore.setPostGameMod(newApostGameMod);

         vScore.setBaseScore(newVbaseScore);
         vScore.setPostGameMod(newVpostGameMod);
      } catch (AssassinSQLException asqle) {
			if(Debug.debug) Debug.print(this, "Exception while scoring deaths: "+asqle);
         return;
      }
		if(Debug.debug) Debug.print(this, "Done scoring death.");
   }
}


