/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.strawberry;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ucam.srcf.assassins.Assassin;
import org.ucam.srcf.assassins.AssassinActivation;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.Event;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.PluginConnector;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.Resurrection;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.Timer;
import org.ucam.srcf.assassins.Tools;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class StrawberryStatic
{
   Strawberry strawberry;
   PluginConnector vanilla;
   String scoreField = "BaseScore";
   String innocentField = "Innoccents";
   String bonusField = "ModBonus";

   void setPluginConnector(PluginConnector p) {
      vanilla=p;
   }

   StrawberryStatic(Strawberry st, PluginConnector vanilla) {
      strawberry=st;
      this.vanilla=vanilla;
   }

   void staticScore(Map params) throws AssassinSQLException {
      Map scores = staticScoreTill(strawberry.startTime, vanilla.getNowTime(), false);
      if(scores == null)
         return;
      
      /* Print the answers */
      List assassins=Assassin.getAllAssassinViewers();
      Iterator i=assassins.iterator();
      while(i.hasNext()) {
         AssassinViewer current=(AssassinViewer) i.next();
         Double score = (Double) scores.get(current);
         if(score==null)
            score=new Double(strawberry.defaultStartingBaseScore);
         System.out.println(score.doubleValue()+" : "+current.getPlayerViewer().getName());
      }
   }
   
   /**
    * Scores till the given time.
    * from the game start time
    * @param from    Misnamed really. The time to start fixing scores
    * @param till    The time to score till
    * @param check   If true then after each event score it and 
    *                   check the answers are correct in {from,till}.
    * @return A Map from AssassinViewers to scores
    */
   Map staticScoreTill(long from, long till, boolean check) throws AssassinSQLException {
      Timer scoreTimer = new Timer("Getting AssassinViewers");
      /* Get a list of all assassins */
      List assassins=Assassin.getAllAssassinViewers();
      scoreTimer.debugPrintTimer();

      scoreTimer = new Timer("Initializing");
      /* Create our scoring object */
      String algName = Config.getConfig(strawberry.getPluginName(), "StaticScorer");
      StaticScoringAlgorithm statScorer;
      try {
         Class algCls = Class.forName(algName);
         statScorer = (StaticScoringAlgorithm) algCls.newInstance();
      } catch (Exception ex) {
         System.out.println("Error loading class "+algName);
         System.out.println("Aborting.");
         return null;
      }
      statScorer.init(assassins.size(), strawberry);
      Map masterScores = null;
      if(check) {
         masterScores = new HashMap();
         Iterator i = assassins.iterator();
         while(i.hasNext())
            masterScores.put(i.next(), new Double(strawberry.defaultStartingBaseScore));
      }
      scoreTimer.debugPrintTimer();

      /* Build state */
      scoreTimer = new Timer("Building State");
      Set dead=new HashSet();
      if(Debug.debug) Debug.print(StrawberryStatic.class, "Building state");
      List sclist = StateChange.getVitalityActivationAndPluginChangesBetweenTimes(strawberry.startTime, till, strawberry.getPluginName());
      Event.preCacheSubClasses(Event.READALL);
      Iterator i = sclist.iterator();
      if(Debug.debug) Debug.print(StrawberryStatic.class, "# statechanges in time ("+strawberry.startTime+","+till+") = "+sclist.size());
      EventViewer last_ev = null;
      List score_changes = null;
      boolean changed = false;
      long time = 0;
      boolean assAct = false;
      while(i.hasNext()) { 
         StateChange sc = (StateChange) i.next();
         if (null==sc) {
            if (Debug.debug) Debug.print(StrawberryStatic.class, "Danger Will Robinson! State Change is null!!");
            continue;
         }
         EventViewer e = sc.getEventViewer();
         if (null==e) {
            if (Debug.debug) Debug.print(StrawberryStatic.class, "Danger Will Robinson! Event for StateChange is null!! ("+sc+")");
            continue;
         }

         if(Debug.debug) Debug.print(StrawberryStatic.class, "Processing StateChange "+sc+" on event ("+e+") \""+e.getHeadline()+"\" at time "+Tools.prettyDateAndTime(e.getTime()));

         if(!e.equals(last_ev)) {
            if(null != last_ev && check && !assAct && time >= from) {
               if(Debug.debug) Debug.print(StrawberryStatic.class, "Updating scores on event "+last_ev);
               updateStaticScoresOnEvent(last_ev, changed, score_changes, masterScores, statScorer);
            } else if(assAct) {
               /* we consdier the event authorative */
               Iterator j = score_changes.iterator();
               while(j.hasNext()) {
                  StateChange scc = (StateChange) j.next();
                  masterScores.put(scc.getAssassinViewer(), new Double((String) scc.getValue()));
               }
            }

            score_changes = new ArrayList();
            changed = false;
            time = e.getTime();
            last_ev = e;
            assAct = false;

            if(Debug.debug) Debug.print(StrawberryStatic.class, "New Event start: event ("+e+") \""+e.getHeadline()+"\" at time "+Tools.prettyDateAndTime(e.getTime()));
         }

         if(sc instanceof AssassinActivation) {
            assAct=true;
            if(Debug.debug) Debug.print(this, "Marking event as assassin activation so should be ignored for correctional purposes");
         }
         
         if(sc instanceof Death && !dead.contains(((Death) sc).getVictim())) {
            if(Debug.debug) Debug.print(StrawberryStatic.class, "Death of someone who isn't dead yet ("+((Death) sc).getVictim().getPlayerViewer().getName()+"), scoring");
            statScorer.addDeath((Death) sc);
            dead.add(((Death) sc).getVictim());
            /* The event has something to write home about */
            changed = true;
         }
         
         if(sc instanceof Resurrection) {
            if(Debug.debug) Debug.print(StrawberryStatic.class, "Ressurrection of "+sc.getAssassinViewer());
            dead.remove(sc.getAssassinViewer());
         }

         /* Only PluginStateChanges beyond this point */
         if(!(sc instanceof PluginStateChange))
            continue;
         
         String field=sc.getFieldName();

         if(innocentField.equals(field)) {
            statScorer.reportInnocents(sc.getAssassinViewer(), Integer.parseInt((String) sc.getValue()));
            changed = true;
         }

         if(bonusField.equals(field)) {
            statScorer.giveBonus(sc.getAssassinViewer(), Double.parseDouble((String) sc.getValue()));
            changed = true;
         }
         
         if(check) {
            if(scoreField.equals(field)) {
               if(time >= from && !assAct) {
                  if(Debug.debug) Debug.print(StrawberryStatic.class, "Score change of "+sc.getAssassinViewer()+", adding to list to check");
                  score_changes.add(sc);
               } else {
                  if(Debug.debug) Debug.print(StrawberryStatic.class, "Score change of "+sc.getAssassinViewer()+", considering authoritative");
                  masterScores.put(sc.getAssassinViewer(), new Double((String) sc.getValue()));
               } // if time >= from && !assAct
            } // if scoreField == field
         } // if check
      } // while i.hasNext
      if(Debug.debug) Debug.print(StrawberryStatic.class, "Debugging Timer");
      scoreTimer.debugPrintTimer();

      /* Score */
      scoreTimer = new Timer("Scoring");
      if(Debug.debug) Debug.print(StrawberryStatic.class, "Scoring");
      Map scores=statScorer.score();
      scoreTimer.debugPrintTimer();

      return scores;
   }

   private void updateStaticScoresOnEvent(EventViewer e, boolean changed, List score_changes, Map masterScores, StaticScoringAlgorithm statScorer) throws AssassinSQLException {
      /* if the scores didn't change, remove and scorechanges and be done*/
      if(!changed) {
         if(Debug.debug) Debug.print(StrawberryStatic.class, "Nothing changed, removing all score related SCs from event");
         Iterator j = score_changes.iterator();
         while(j.hasNext()) {
            StateChange sc = (StateChange) j.next();
            vanilla.removeStateChange(e, sc);
         }
      } else {
         /* First process preexisting scores, then add anyone who got missed */
         if(Debug.debug) Debug.print(StrawberryStatic.class, "Rescoring...");
         
         /* Get the new scores */
         Map scores = statScorer.score();
         
         /* Set of people we need to write scores for */
         Set todo = scores.keySet(); 
         Set done = new HashSet();
         
         /* Iterate over the pre-existing score changes on this event */
         if(Debug.debug) Debug.print(StrawberryStatic.class, "Considering prior score changeds for correctness");
         Iterator j = score_changes.iterator();
         while(j.hasNext()) {
            StateChange sc = (StateChange) j.next();
            AssassinViewer who = sc.getAssassinViewer();
            
            Double new_score = (Double) scores.get(who);
            Double old_score = new Double((String) sc.getValue());
            Double current_score = (Double) masterScores.get(who);

            if(Debug.debug) Debug.print(StrawberryStatic.class, "Checking for "+who.getPlayerViewer().getName()+" old score was "+old_score+" and new score is "+new_score);
            
            if(!scoresDiffer(new_score, current_score)) {
               /* no need for this statchange so remove it */
               if(Debug.debug) Debug.print(StrawberryStatic.class, "new and master scores the same, score change is pointless");
               vanilla.removeStateChange(e, sc);
            } else {
               /* we need a score change on here, si the current one correct? */
               if(scoresDiffer(new_score, old_score)) {
                  /* no its not */
                  if(Debug.debug) Debug.print(StrawberryStatic.class, "Old score is wrong, correcting it.");
                  sc = new PluginStateChange(who, strawberry, scoreField, ""+new_score.intValue());
                  vanilla.addStateChange(e, sc);
               }
               /* and update the master score regardless */
               masterScores.put(who, new_score);
            }
            done.add(who);
         }

         /* Iterate over whoever is left */
         if(Debug.debug) Debug.print(StrawberryStatic.class, "Iterating over remaining assassins to check scores for");
         j = todo.iterator();
         while(j.hasNext()) {
            AssassinViewer who = (AssassinViewer) j.next();
            if(done.contains(who))
               continue;
            
            Double new_score = (Double) scores.get(who);
            Double current_score = (Double) masterScores.get(who);
            
            if(Debug.debug) Debug.print(StrawberryStatic.class, "Considering "+who.getPlayerViewer().getName()+" master score "+current_score+" new score "+new_score);
            
            /* Has the score actually changed? */
            if(scoresDiffer(new_score, current_score)) {
                  if(Debug.debug) Debug.print(StrawberryStatic.class, "score differs from masterScore for this player, emitting statechange");
               /* If it has then write out the new one and update the
                * master listing */
               StateChange sc = new PluginStateChange(who, strawberry, scoreField, ""+new_score.intValue());
               vanilla.addStateChange(e, sc);
               masterScores.put(who, new_score);
            }
         }
      }
   }

   private boolean scoresDiffer(Double s1D, Double s2D) {
      double s1d = strawberry.defaultStartingBaseScore;
      if(null != s1D)
         s1d = s1D.doubleValue();
      double s2d = strawberry.defaultStartingBaseScore;
      if(null != s2D)
         s2d = s2D.doubleValue();
      return (Math.abs(s1d - s2d) > 0.0001);
   }
}


