#include <jni.h>
#include <stdlib.h>

#include "ks_spsol.h"

#include "meschach/sparse.h"
#include "meschach/matrix.h"

#include <math.h>

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Initialises memroy for n assassins
 */
extern void Java_org_ucam_srcf_assassins_strawberry_KirikaSparse_realinit (JNIEnv *env, jclass cls, jint narg)
{
   //fprintf(stderr, "Init started\n");
   n = narg;

   kill_matrix = calloc(n*n, sizeof(double));
   if(kill_matrix == NULL)
   {
      throw_out_of_mem_exception(env);
      return;
   }

   death_count = (double *) calloc(n, sizeof(double));
   if(death_count == NULL)
   {
      throw_out_of_mem_exception(env);
      return;
   }

  //fprintf(stderr, "Init successful\n");
}

/*
 * deallocate memory
 */
extern void Java_org_ucam_srcf_assassins_strawberry_KirikaSparse_realfinal (JNIEnv *env, jclass cls)
{
   n=0;
   free(kill_matrix);
   free(death_count);
}
   
/*
 * Adds a death to our matrix
 */
extern void Java_org_ucam_srcf_assassins_strawberry_KirikaSparse_realAddDeath (JNIEnv *env, jclass cls, jint killer, jint victim, jdouble fac)
{
   //fprintf(stderr, "adding death %i -> %i ...", killer, victim);
   if(kill_matrix == NULL)
   {
      throw_non_init_exception(env);
      return;
   }
   if(killer >= n)
   {
      throw_arrayOOB(env, killer);
      return;
   }
   if(victim >= n)
   {
      throw_arrayOOB(env, victim);
      return;
   }
   
   kill_matrix[killer*n+victim]++;

   death_count[victim] += level_entry(kill_matrix[killer*n+victim], fac) -
                          level_entry(kill_matrix[killer*n+victim] - 1, fac); 
   //fprintf(stderr, " done\n");
}

void printMatrix(int n, double *A) {
   int i,j;

   for(i=0;i<n;i++)
   {
      for(j=0;j<n;j++)
         fprintf(stderr, "%g   ", A[n*i+j]);
      fprintf(stderr, "\n");
   }
}

void printVector(int n, double *v) {
   int i;

   for(i=0; i<n; i++)
     fprintf(stderr, "%g   ", v[i]);
  fprintf(stderr, "\n");
}

/*
 * Solve the system using conjugate gradients given guess u for a start point
 * using c as the scoring constant
 */
extern void Java_org_ucam_srcf_assassins_strawberry_KirikaSparse_realsolve (JNIEnv *env, jclass cls, jdoubleArray u, jint c, jdouble tol, jdouble fac)
{
   int i,j;
   double kc;     /*kill constant*/
   double temp;
   SPMAT *A;
   VEC *b, *x, *r0;
   double *u_elems;

   //fprintf(stderr, "Solving... \n");

   /* TODO: Validate input */
   if(kill_matrix == NULL) {
      throw_non_init_exception(env);
      return;
   }
   
   //fprintf(stderr, "kill_matrix:\n");
   //printMatrix(n, kill_matrix);
   //fprintf(stderr, "Death Vector:\n");
   //printVector(n, death_count);
   
   /* alloc memory */
   A = sp_get(n,n,5);      /* 5 is a reaonable number of 
                              non-zeros per row - ie kills */

   /* find the kill constant (c-1)/c */
   kc = (c-1.0)/c;
   
   /* Convert to a sparse matrix while counting deaths
    * after levelling and put -1's down the diagonal as we go
    * form the matrix A by (1-[(c-1)/c]^m)/deaths * kills */
   for(i=0;i<n;i++) 
   {
      /* main inner loop */
      for(j=0;j<n;j++)
      {
         /* Make sure theres an i,i entry in the logical place for speed 
          * of creation */
         if(i==j)
         {
            sp_set_val(A, i, j, -1);
            continue;
         }
         
         /* Skip zeros */
         if(kill_matrix[i*n+j] == 0) continue;

         /* Level it and add it if its non-zero and non-diagonal */
         temp = level_entry(kill_matrix[i*n+j], fac);
         if(death_count[j] != 0) {
            temp *= (1-pow(kc, death_count[j]));
            temp /= death_count[j];
         }
         sp_set_val(A, i, j, temp); 
      }
   }
   
   /* Form the dense vector b as -1 */
   b = v_get(n);
   for(i=0; i<n; i++)
      v_set_val(b, i, -1);

   /* Read the inital guess into vector x */
   x = v_get(n);
   u_elems = (*env)->GetDoubleArrayElements(env, u, 0);
   for(i=0; i<n; i++)
      v_set_val(x, i, u_elems[i]);

   /* produce a random r0 vector */
   r0 = v_get(n);
   v_rand(r0);
   
   /*fprintf(stderr, "A:\n");
   sp_foutput(stderr, A);
   fprintf(stderr, "b:\n");
   v_foutput(stderr, b);*/
   
   /* call the solve routine using the inital point we are given */
   sp_cgs(A, b, r0, tol, x);

   /*fprintf(stderr, "x:\n");
   v_foutput(stderr, x);*/
   
   /* copy answer back in u */
   for(i=0; i<n; i++)
     u_elems[i] =  v_entry(x, i)*pow(kc, death_count[i]);
   (*env)->ReleaseDoubleArrayElements(env, u, u_elems, 0);
   
   /* deallocate temporary vars */
   v_free(b);
   v_free(r0);
   v_free(x);
   sp_free(A);

   /*fprintf(stderr, "done\n");*/
}

/*
 * Level a matrix entry
 */
double level_entry(double entry, double fac) 
{
   double ret;
   int i;
   
   for(i=1, ret=0; i<=entry; i++)
   {
      if(i<=2)
         ret+=1;
      else if(i>=7)
         ret+=0.1;
      else
         ret+=1-fac*(i-2);
   }

   return ret;
}

/*
 * Throws a Java exception as we haven't initialised memory yet
 * and we are trying to do something with it
 */
void throw_non_init_exception(JNIEnv *env)
{
	jclass newExcCls;
	newExcCls = (*env)->FindClass(env, "java/lang/RuntimeException");
	(*env)->ThrowNew(env, newExcCls, "C code not initilaised yet.");
}

/*
 * Throws a Java exception if we have difficulty allocating memory
 */
void throw_out_of_mem_exception(JNIEnv *env)
{
	jclass newExcCls;
	newExcCls = (*env)->FindClass(env, "java/lang/RuntimeException");
	(*env)->ThrowNew(env, newExcCls, "C Code out of Memory.");
}

/*
 * Throws a Java exception if we are trying to access past the end 
 * of an array at index index.
 */
void throw_arrayOOB(JNIEnv *env, int index)
{
	jclass newExcCls;
	newExcCls = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
	(*env)->ThrowNew(env, newExcCls, "C code called to access past end of array.");
}

#ifdef __cplusplus
}
#endif

