/*
 * CUAG AutoUmpire System
 *
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.watermelon;

import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.PlayerViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.TextQuestion;
import org.ucam.srcf.assassins.YesNoQuestion;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class Watermelon extends Plugin
{
   public String getDescription() { return "Set a `wanted' flag on assassins."; };
   /**
    * Called when the game is first created, to allow the plugin to do any
    * initialisation necessary.
    */
   protected void gameCreated() {}


   /**
    * Called whenever the plugin is required to update or sanity-check the
    * current gamestate.
    */
   protected Vector updateState(Map state, Vector qs, EventViewer e) throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(Watermelon.class, "updateState("+state+","+qs+")");
      Vector v = new Vector();
      Iterator i = state.keySet().iterator();
      int j = 0;
      while(i.hasNext()) {
         AssassinViewer a = (AssassinViewer) i.next();
         PlayerViewer p = a.getPlayerViewer();
         AssassinStateViewer asv = (AssassinStateViewer) state.get(a);
         boolean wanted = asv.getBoolProperty(this, "wanted");
         if (Debug.debug) Debug.print(Watermelon.class,"Wanted?: " + asv.getProperty(this, "wanted"));
         if (wanted) {
            if (0 != qs.size()) {
               YesNoQuestion q = (YesNoQuestion) qs.get(j++);
               if (Debug.debug) Debug.print(Watermelon.class, "Got answer: " + q.getAnswer());
               if (q.getAnswer()) {
                  vanilla.addStateChange(e, new PluginStateChange(a, this, "wanted", "false"));
               }
            }
            else {
               String redemption=asv.getProperty(this,"redemption");
               v.add(new YesNoQuestion("Redeem "+p.getName()+"? ("+redemption+")", false));
            }
         }
      }
      if (0 == qs.size()) return null;
      return v;
   }

   /**
    * Called by the filtering mechanism when the plugin needs to determine
    * whether an assassin has a given property such as &quot;police&quot;. If
    * the plugin does not recognise the named property, it should return
    * <code>false</code>.
    * @param a A viewer for the assassin being tested.
    * @param prop The name of the property being asked about.
    * @return True if the assassin has the property, false if not.
    */
   public  boolean assassinHasProperty(AssassinStateViewer asv, String prop, String value)
   {
      if (Debug.debug) Debug.print(Watermelon.class, "assassinHasProperty("+asv+","+prop+","+value+")");
      if (!"wanted".equals(prop)) return false;
      String wanted = asv.getProperty(this, "wanted");
      if (null == wanted) return "false".equals(value);
      if (wanted.equals(value)) return true;
      return false;
   }

   /**
    * Called during the updating of the gamestate to ask whether a given player
    * requires email to be sent to them during the current round of emails.
    * @param p A viewer for the player being asked about.
    * @return True if the plugin requires email to be sent to the player in
    *    question, false if not.
    */
   protected boolean playerNeedsEmail(AssassinStateViewer asv)
   {
      return false;
   }

   /**
    * Called when the system has decided that email needs to be sent to a given
    * player. The plugin may return a string to be included in the email, or
    * null if it does not wish to include anything.
    * @param p A viewer for the player being emailed.
    * @return A string to be included in the email sent to the player, or null.
    *    The string will appear in its own paragraph when inserted into the
    *    email and so should not begin or end with a newline.
    */
   protected String generateEmail(AssassinStateViewer asv)
   {
      return "";
   }

   /**
    * Called when an event is being added to get any queries the plugin
    * has for the user. The plugin should check all the event details and if it
    * needs more information it should ask relevent questions. The
    * returnQuestions function is passed the same Vector with the answers
    * filled out.
    * @param e The Event which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector participantAdded(EventViewer e, AssassinViewer a, Map state, Vector qs) throws AssassinSQLException
   {
      if (null == qs) {
         if (Debug.debug) Debug.print(Watermelon.class, "getQuestions("+e+","+state+")");
         qs = new Vector();
         PlayerViewer p = a.getPlayerViewer();
         qs.add(new YesNoQuestion("Make " + p.getName() + " wanted?", false));
         qs.add(new TextQuestion("Crime:", ""));
         qs.add(new TextQuestion("Redemption Conditions:", ""));
         return qs;
      } else {
         if (Debug.debug) Debug.print(Watermelon.class, "returnQuestions("+e+","+state+","+qs+")");
         boolean wanted = ((YesNoQuestion) qs.get(0)).getAnswer();
         String crime = ((TextQuestion) qs.get(1)).getAnswer();
         String redemption = ((TextQuestion) qs.get(2)).getAnswer();
         if (wanted) {
            vanilla.addStateChange(e, new PluginStateChange(a, this, "wanted", "true"));
            vanilla.addStateChange(e, new PluginStateChange(a, this, "crime", crime));
            vanilla.addStateChange(e, new PluginStateChange(a, this, "redemption", redemption));
         }
         return null;
      }
   }

}
